# import lib
import numpy as np
import sys
sys.path.append("..")
from iimcsim.MonteCarlo import MonteCarlo as mc
from tqdm import tqdm

class MC_set:
    def __init__(self, filepath, max_Rate, exe, sample_size):
        self.filepath = np.load(filepath)
        self.max_Rate= max_Rate
        self.exe = exe
        self.sample_size = sample_size
    def remove_duplicate_pulse(self):
        if len(self.filepath.shape)<2:
            x_all = self.filepath.reshape(int(self.filepath.shape[0]/100),100)
        else:
            x_all = self.filepath.copy()
        # Remove duplicates
        shapes = [x_all[i][x_all[i]!=0] for i in range(x_all.shape[0])]# creates list of shapes
        uniq_shapes = [list(j) for j in set(map(tuple,shapes))]# sublist match in the form of tuple
        print(f"Total unique shapes:{len(uniq_shapes)}")
        return uniq_shapes
    def train_test_val_split(self, train, test):
        y = self.remove_duplicate_pulse()
        train_lim = int(len(y)*(train/100))
        test_lim = int(len(y)*(test/100))
        #val_lim = int(len(y)*(val/100))
        return y[:train_lim], y[train_lim:]#, y[train_lim+test_lim:]

    def get_mc_set(self, shape_list):
        data_X = []; data_X1 = []
        data_y = []; data_y1 = []
        CORR_pattern = []
        CH1_ind = []
        CH2_ind = []
        for i in tqdm(range(self.max_Rate)):
            x, y, x1, y1, cor_pat, ch1_ind, ch2_ind = mc(shape_list, i, self.exe, self.sample_size).MC2()
            data_X.append(x); data_X1.append(x1)
            data_y.append(y); data_y1.append(y1)
            CORR_pattern.append(cor_pat)
            CH1_ind.append(ch1_ind)
            CH2_ind.append(ch2_ind)
        #data_X, data_y = mc(shape_list,2,10,100).MC2()
        data_X = np.array(data_X); data_X = data_X.reshape(self.exe*self.max_Rate, self.sample_size)
        data_y = np.array(data_y); data_y = data_y.reshape(self.exe*self.max_Rate, self.sample_size)

        data_X1 = np.array(data_X1); data_X1 = data_X1.reshape(self.exe*self.max_Rate, self.sample_size)
        data_y1 = np.array(data_y1); data_y1 = data_y1.reshape(self.exe*self.max_Rate, self.sample_size)
        return data_X, data_y, data_X1, data_y1, CORR_pattern, CH1_ind, CH2_ind



import numpy as np
np.random.seed(42)
from scipy.signal import find_peaks
import matplotlib.pyplot as plt


class MonteCarlo:

    def __init__(self, input_uniq_shapes, photon_per_samp, required_examples, sample_size, samp_swap=True):
        self.input_uniq_shapes = input_uniq_shapes
        self.photon_per_samp = photon_per_samp
        self.required_examples = required_examples
        self.sample_size = sample_size
        self.shape_lengths = np.array([len(shape) for shape in input_uniq_shapes])
        self.samp_swap = samp_swap

    def remove_zeros(self, arr):
        return np.array(arr)[np.array(arr) != 0]

    def get_samp_from_ind(self, ind, shape_samples):
        # print(shape_samples)
        samples = [shape_samples[i] for i in ind]
        return samples

    def convert_samples_to_only_shapes(self, samples):
        shapes = [self.remove_zeros(i) for i in samples]
        return shapes

    def samplewise_shape_list_reshape(self, shape):
        reshaped_arr = np.array(shape).reshape(self.required_examples, self.photon_per_samp)
        return reshaped_arr

    def upper_lower_bounds(self, shape, peak_context=True):
        peak, _ = find_peaks(shape)
        shape_length = len(shape)
        if peak_context is False:
            if len(peak) == 0 and shape_length == 1:
                # print("Spike")
                lower_bound = 0
                upper_bound = self.sample_size - 1
            elif len(peak) == 0 and shape_length > 1:
                # print("Plateau")
                lower_bound = 0 - shape_length + 1
                upper_bound = 255 + shape_length - 1
            elif len(peak) != 0 and shape_length > 1:
                # print("Proper pulse")
                lower_bound = 0 - shape_length + 1
                upper_bound = 255 + shape_length - 1
            return lower_bound, upper_bound
        elif peak_context is True:
            if len(peak) == 0 and shape_length == 1:
                # print("Spike")
                lower_bound_pk = 0
                upper_bound_pk = self.sample_size - 1

                len_before_peak = 0
                len_after_peak = 0
                before_peak_part = 0
                after_peak_part = 0
            elif len(peak) == 0 and shape_length > 1:
                # print("Plateau")
                lower_bound_pk = 0 - shape_length + 1
                upper_bound_pk = self.sample_size - 1

                len_before_peak = 0
                len_after_peak = len(shape)
                before_peak_part = 0
                after_peak_part = shape
            elif len(peak) != 0 and shape_length > 1:
                # print("Proper pulse")
                before_peak_part = shape[:peak[0]]
                after_peak_part = shape[peak[0]:]

                len_before_peak = len(before_peak_part)
                len_after_peak = len(after_peak_part)

                lower_bound_pk = 0 - len_after_peak + 1
                upper_bound_pk = self.sample_size + len_before_peak - 1
            return lower_bound_pk, upper_bound_pk, len_before_peak, len_after_peak, before_peak_part, after_peak_part

    def get_ind_for_pulse_to_add(self, lower_bound, upper_bound):
        indice = np.random.randint(lower_bound, upper_bound, 1)
        return indice[0]

    def conditions_to_add_shapes(self, ind, shape, X, Y):
        if len(shape)==1:
            # single spike
            X[ind:ind+1]+=shape
            Y[ind]+=1
        else:
            peak_list, _ = find_peaks(shape)
            if not peak_list:
                full_plat_l_end = ind
                full_plat_h_end = ind + len(shape)
                if full_plat_l_end == 0:
                    X[:full_plat_h_end] += shape
                    Y[0] += 1
                elif full_plat_h_end == self.sample_size:
                    X[full_plat_l_end:] += shape
                    Y[full_plat_l_end] += 1
                elif full_plat_l_end>0 and full_plat_h_end<self.sample_size:
                    X[full_plat_l_end:full_plat_h_end] += shape
                    Y[full_plat_l_end] += 1
                elif full_plat_l_end<0 and full_plat_h_end>=0:
                    part_to_add = shape[abs(full_plat_l_end):]
                    X[:len(part_to_add)] += part_to_add
                    Y[0] += np.trapz(part_to_add)/np.trapz(shape)
                elif full_plat_l_end>0 and full_plat_h_end>self.sample_size:
                    part_to_add = shape[:abs(full_plat_l_end-self.sample_size)]
                    X[full_plat_l_end:] += part_to_add
                else:
                    print('----------Reconsider the MC simulation---------------------')
                    print(ind, shape)
                    print(full_plat_l_end, full_plat_h_end)
                    print('----------Reconsider the MC simulation---------------------')
                    
            else:
                # Pulse without plateau
                full_pulse_peak_ind = peak_list[0]
                full_pulse_l_end = ind-full_pulse_peak_ind
                full_pulse_h_end = full_pulse_l_end + len(shape)
                if ind==0: # peak at 0
                    part_to_add = shape[full_pulse_peak_ind:]
                    X[:len(part_to_add)] += part_to_add
                    Y[0] += np.trapz(part_to_add)/np.trapz(shape)
                elif ind==self.sample_size-1: # peak at end e.g, 255
                    part_to_add = shape[:full_pulse_peak_ind+1]
                    X[(self.sample_size-1)-full_pulse_peak_ind:] += part_to_add
                    Y[self.sample_size-1] += np.trapz(part_to_add)/np.trapz(shape)
                    
                elif ind<0: # peak beyond 0(negative)
                    part_to_add = shape[abs(full_pulse_l_end):]
                    X[:full_pulse_h_end] += part_to_add
                    Y[0] += np.trapz(part_to_add)/np.trapz(shape)
                elif ind>255: # peak beyond 255
                    part_to_add = shape[:(self.sample_size-full_pulse_l_end)]
                    X[full_pulse_l_end:] += part_to_add
                    fraction_of_pulse = np.trapz(part_to_add)/np.trapz(shape)
                    Y[255] += fraction_of_pulse
                elif ind>0:
                    if full_pulse_l_end>0 and full_pulse_h_end<self.sample_size:
                        X[full_pulse_l_end:full_pulse_h_end] += shape
                        Y[ind] += 1
                    elif full_pulse_l_end<0 and full_pulse_h_end<self.sample_size:
                        part_to_add = shape[abs(ind):]
                        X[:len(part_to_add)] += part_to_add
                        Y[ind] += np.trapz(part_to_add)/np.trapz(shape)
                    elif full_pulse_l_end>0 and full_pulse_h_end>self.sample_size:
                        part_to_add = shape[:abs(full_pulse_l_end-self.sample_size)]
                        X[full_pulse_l_end:] += part_to_add
                        Y[ind] += np.trapz(part_to_add)/np.trapz(shape)
                    elif full_pulse_l_end == 0 and full_pulse_h_end<self.sample_size:
                        X[:len(shape)] += shape
                        Y[ind] += 1
                    else: #full_pulse_l_end>0 and full_pulse_h_end==self.sample_size-1:
                        X[full_pulse_l_end:] += shape
                        Y[ind] += 1
                        
                
        
        
    def shapes_according_to_samples(self):
        samples_to_shapes = self.convert_samples_to_only_shapes(self.input_uniq_shapes)
        select_shape_ind_from_uniq_shapes = np.random.randint(len(self.input_uniq_shapes), size=(
        self.required_examples, self.photon_per_samp)).astype(int)
        select_shape_ind_from_uniq_shapes1 = np.random.randint(len(self.input_uniq_shapes), size=(
        self.required_examples, self.photon_per_samp)).astype(int)
        selected_shapes_to_be_added_in_samples = [
            samples_to_shapes[k1]
            for k in select_shape_ind_from_uniq_shapes
            for k1 in k
        ]
        selected_shapes_to_be_added_in_samples1 = [
            samples_to_shapes[k01]
            for k0 in select_shape_ind_from_uniq_shapes1
            for k01 in k0
        ]
        samplewise_reshaped_shapelist = self.samplewise_shape_list_reshape(selected_shapes_to_be_added_in_samples)
        samplewise_reshaped_shapelist1 = self.samplewise_shape_list_reshape(selected_shapes_to_be_added_in_samples1)
        return samplewise_reshaped_shapelist, samplewise_reshaped_shapelist1

    def correlation_manipulation_condition(self, j2,
                                           lb, ub, shape,
                                           lb1, ub1, shape1):
        get_ind_randomly_for_pulse_peak_to_add = self.get_ind_for_pulse_to_add(lb, ub)
        get_ind_randomly_for_pulse_peak_to_add1 = self.get_ind_for_pulse_to_add(lb1, ub1)
        
        if j2 == 1 and get_ind_randomly_for_pulse_peak_to_add<self.sample_size-1 and get_ind_randomly_for_pulse_peak_to_add>0:
            get_ind_randomly_for_pulse_peak_to_add1 = get_ind_randomly_for_pulse_peak_to_add
            if self.samp_swap is True:
                lb1 = lb;
                ub1 = ub;
                shape1 = shape
            else:
                pass
        elif j2 == 0:
            get_ind_randomly_for_pulse_peak_to_add = self.get_ind_for_pulse_to_add(lb, ub)
            get_ind_randomly_for_pulse_peak_to_add1 = self.get_ind_for_pulse_to_add(lb1, ub1)
        return get_ind_randomly_for_pulse_peak_to_add, get_ind_randomly_for_pulse_peak_to_add1, shape, shape1

    
    def MC2(self):
        arr_x = np.zeros((self.required_examples, self.sample_size))
        arr_y = np.zeros((self.required_examples, self.sample_size))
        arr_x1 = arr_x.copy()
        arr_y1 = arr_y.copy()
        
        ch1_ch2_corr_pattern = []
        ch1_pulse_peak_ind = []
        ch2_pulse_peak_ind = []
        
        samplewise_reshaped_shapelist, samplewise_reshaped_shapelist1 = self.shapes_according_to_samples()
        for i, X, Y, X1, Y1 in zip(range(self.required_examples), arr_x, arr_y, arr_x1, arr_y1):
            # This for loop gives an array of shapes which are selected to be added in a sample data
            corr_arr = np.random.choice([0, 1], size=self.photon_per_samp, replace=True)
            pulse_add_ind = []
            pulse_add_ind1= []
            for j, j1, j2 in zip(samplewise_reshaped_shapelist[i], samplewise_reshaped_shapelist1[i], corr_arr):
                # This loop is for accessing those shapes one by one
                # here j is the pulse
                lb, ub, len_bp, len_ap, bp_part, ap_part = self.upper_lower_bounds(j)
                lb1, ub1, len_bp1, len_ap1, bp_part1, ap_part1 = self.upper_lower_bounds(j1)
                get_ind_randomly_for_pulse_peak_to_add, get_ind_randomly_for_pulse_peak_to_add1, shape, shape1 = self.correlation_manipulation_condition( j2, lb, ub, j, lb1, ub1, j1)
                self.conditions_to_add_shapes(get_ind_randomly_for_pulse_peak_to_add, shape, X, Y)
                self.conditions_to_add_shapes(get_ind_randomly_for_pulse_peak_to_add1, shape1, X1, Y1)
                pulse_add_ind.append(get_ind_randomly_for_pulse_peak_to_add)
                pulse_add_ind1.append(get_ind_randomly_for_pulse_peak_to_add1)
            ch1_ch2_corr_pattern.append(corr_arr)
            ch1_pulse_peak_ind.append(pulse_add_ind)
            ch2_pulse_peak_ind.append(pulse_add_ind1)
            
        return arr_x, arr_y, arr_x1, arr_y1, ch1_ch2_corr_pattern, ch1_pulse_peak_ind, ch2_pulse_peak_ind


#input_uniq_shapes = np.load('./src/data/shaula_00000-003_ch0_shapes.npy')
#photon_per_samp = 15
#required_examples = 100
#sample_size = 256

#q = MonteCarlo(input_uniq_shapes, photon_per_samp, required_examples, sample_size, samp_swap=True)
#x, y, x1, y1, cor_pat, ch1_ind, ch2_ind = q.MC2()



import iimcsim.tools as fg
import numpy as np
from iimcsim.shape_ext import shape_generator


def shapes_extraction(calib_data_path='../src/data/calib_ch0.npy',
                      high_rate_path = '../src/data/shaula_00000-003_ch0.npy',
                      data_save_path='../src/data/',
                      bins = 21400
                      ):
    # required files
    # calib.npy and shaula_00000-003_ch0.npy
    select_files = shape_generator(calib_data_path, high_rate_path,data_save_path)
    generate_shapes = select_files.generator(bin_num=bins)
    shapes, save_path = select_files.scalar(generate_shapes)
    # generated file
    # shaula_00000-003_ch0_shapes.npy
    return shapes

def pulse_sh_plat(shapes_path = '../src/data/shaula_00000-003_ch0_shapes.npy',
                  pulse_info_path = '../src/data/',
                  plat_size_sel = 1,
                  plot_save_path = '../plots/',
                  lst=[3,4,5]):
    uniq_shape = np.load(shapes_path)
    # Generate data of extracted pulse, individually; pulse height, pulse width, skewness, kurtosis in dataframe format
    df = fg.pulse_data(uniq_shape)
    df.to_csv('calib_pulse_data.csv')
    fg.skew_kurt(df, plat_size_sel, plot_save_path)
    platauless_uniq_shapes, uniq_shapes = fg.filter_shapes_based_on_plat_size(df, uniq_shape, platau_removal=True, lst=lst, save_path=pulse_info_path)
    return platauless_uniq_shapes, uniq_shapes


shapes = shapes_extraction(r"C:\Users\jigar\Downloads\photon_position_reconstruction-main\photon_position_reconstruction-main\src\data\calib_ch0.npy",
                           r"C:\Users\jigar\Downloads\photon_position_reconstruction-main\photon_position_reconstruction-main\src\data\shaula_00000-003_ch0.npy",
                           r"C:\Users\jigar\Downloads\photon_position_reconstruction-main\photon_position_reconstruction-main\src\data")

pulse_sh_plat(r"C:\Users\jigar\Downloads\photon_position_reconstruction-main\photon_position_reconstruction-main\src\data\shaula_00000-003_ch0_shapes.npy",
              r"C:\Users\jigar\Downloads\photon_position_reconstruction-main\photon_position_reconstruction-main\src\data",
              plot_save_path = './')


import pathlib
import numpy as np
import pandas as pd
from scipy.signal import find_peaks

class shape_generator:
    def __init__(self, calib_file, high_rate_file, path_to_save=pathlib.Path().resolve()):
        self.calib_file = calib_file
        self.high_rate_file = high_rate_file
        self.path_to_save = path_to_save
        
    def generator(self, threshold=0, bin_num=2147483600):
        arr0 = -np.load(self.calib_file)[:bin_num] + threshold
        arr = np.where(arr0<0,0,arr0)
        arr_sub = arr-1
        arr_threshold = np.where(arr_sub<0,0,arr_sub)#/np.max(arr_sub)
        #print('calib_shape:',arr_w.shape)
        #print('arr:',arr.shape)
        #print('arr_sub_shape : ', arr_sub.shape)
        #arr_threshold = arr_w.reshape(np.int(arr_w.shape[0]/100),100)

        rate = []
        ind = []
        val = []
        for i in range(arr_threshold.shape[0]):
            peak_val, peak_ind = find_peaks(arr_threshold[i], height=0)
            rate.append(len(peak_val))
            val.append(peak_val)
            ind.append(peak_ind)

        rate = np.asarray(rate)
        ind = np.asarray(ind)
        val = np.asarray(val)
        df = pd.DataFrame(np.hstack((rate,ind)))

        first_non_zero_ind = np.argwhere(arr_threshold[:,0]!=0)
        mod_arr1 = np.delete(arr_threshold,first_non_zero_ind,0)
        last_non_zero_ind = np.argwhere(mod_arr1[:,99]!=0)
        mod_arr2 = np.delete(mod_arr1,last_non_zero_ind,0)

        rate_side = []
        ind_side = []
        val_side = []

        for i in range(mod_arr2.shape[0]):
            peak_ind, peak_val = find_peaks(mod_arr2[i], height=0)
            rate_side.append(len(peak_ind))
            ind_side.append(peak_ind)
            val_side.append(peak_val)

        rate_side = np.asarray(rate_side)
        ind_side = np.asarray(ind_side)
        val_side = np.asarray(val_side)
        #df1 = pd.DataFrame(np.hstack((rate_side,ind_side)))

        rate_side_ind = np.argwhere(rate_side==1) # indices where there is only 1 photon in rate_side samples
        x = mod_arr2[rate_side_ind,:] # using row index create matrix whose samples have only 1 photon
        x=x.reshape(x.shape[0],100) # reshape the matrix
        return x
    
    def scalar(self,x, threshold=0, bin_num=2147483600):
        high_rate = -np.load(self.high_rate_file)[:bin_num] + threshold#[:2147483600]
        x_scale = x/1#np.max(high_rate)
        name = str(self.path_to_save)+self.high_rate_file[-24:].split('.')[0]+'_shapes.npy'
        np.save(name,x_scale)
        return x, name


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.signal import find_peaks
import pickle
import seaborn as sns
import random
from scipy.stats import skew, kurtosis
import warnings
warnings.filterwarnings('ignore')
import tqdm
import matplotlib.ticker as mtick
import glob
from tensorflow import keras
from sklearn.metrics import confusion_matrix
from sklearn import metrics





def pulse_height(uniq_shapes):
    H = [np.max(np.array(uniq_shapes[i])) for i in range(len(uniq_shapes))]
    sns.distplot(H, bins=30, kde=True, color='black')
    plt.title('Pulse Height')
    plt.savefig('pulse_height_platauless_shapes.png')
    plt.show()

def pulse_width(uniq_shapes):
    H = [len(np.array(uniq_shapes[i])[np.nonzero(np.array(uniq_shapes[i]))]) for i in range(len(uniq_shapes))]
    sns.distplot(H, bins=30, kde=True, color='black')
    plt.savefig('pulse_width_platauless_shapes.png')
    plt.title('Pulse Width')
    plt.show()

def pulse_skewness(uniq_shapes):
    H = [skew(np.array(uniq_shapes[i])) for i in range(len(uniq_shapes))]
    ax = sns.distplot(H, bins=30, kde=True, color='black')
    kde_x, kde_y = ax.lines[0].get_data()
    #plotting the two lines
    p1 = plt.axvline(x=-0.5,color='#EF9A9A', alpha=0.01)
    p2 = plt.axvline(x=0.5,color='#EF9A9A', alpha=0.01)
    p3 = plt.axvline(x=1,color='#EF9A9A', alpha=0.01)


    #ax.fill_between(kde_x, kde_y, where=(kde_x<0) | (kde_x>1), 
    #                interpolate=True, color='red', alpha= 0.5)
    #plt.yscale('log')
    plt.title('Pulse Skewness')
    plt.show()
    
def pulse_kurtosis(uniq_shapes):
    H = [kurtosis(np.array(uniq_shapes[i])) for i in range(len(uniq_shapes))]
    ax = sns.distplot(H, bins=30, kde=True, color='black')
    kde_x, kde_y = ax.lines[0].get_data()
    #plotting the two lines
    p1 = plt.axvline(x=-0.5,color='#EF9A9A', alpha=0.01)
    p2 = plt.axvline(x=0.5,color='#EF9A9A', alpha=0.01)
    p3 = plt.axvline(x=1,color='#EF9A9A', alpha=0.01)


    #ax.fill_between(kde_x, kde_y, where=(kde_x<0) | (kde_x>1), 
    #                interpolate=True, color='red', alpha= 0.5)
    #plt.yscale('log')
    plt.title('Pulse Kurtosis')
    plt.show()

def pulse_data(uniq_shape):
    '''
    This function creates a csv file wich contains different parameters
    of the pulses from calibration data
    This includes plateau size, pulse height and width, skweness and kurtosis
    It takes shapes extracted from calibration measurements in the format of
    numpy array for example shape=(100,100)
    '''
    puls_idx = []; plat_s = []; puls_h = []; puls_W = []; skw = []; krt = []
    for i in range(len(uniq_shape)):
        #print(uniq_shape[i])
        peaks, peak_plateaus = find_peaks(uniq_shape[i], plateau_size=0)
        plat_size = peak_plateaus['plateau_sizes'][0]
        pulse_height = np.max(np.array(uniq_shape[i]))
        pulse_width = len(np.array(uniq_shape[i])[np.nonzero(np.array(uniq_shape[i]))])
        skewness = skew(uniq_shape[i])
        kurt = kurtosis(uniq_shape[i])
        puls_idx.append(i)
        plat_s.append(plat_size)
        puls_h.append(pulse_height)
        puls_W.append(pulse_width)
        skw.append(skewness)
        krt.append(kurt)
    df = pd.DataFrame({'pulse_idx': puls_idx,'plateau_size': plat_s,'pulse_height': puls_h,'pulse_width': puls_W,'skewness': skw,'kurtosis': krt})
    return df




def plat_plot(df, bins=50, img_save_path=None):
    fig, axs = plt.subplots(3, 2,figsize=(13,7))
    axs[0, 0].hist(df.loc[df['plateau_size']==1]['pulse_height'], bins=bins)
    axs[0, 0].set_title('Plateau size = 1')
    axs[0, 1].hist(df.loc[df['plateau_size']==2]['pulse_height'], bins=bins)
    axs[0, 1].set_title('Plateau size = 2')
    axs[1, 0].hist(df.loc[df['plateau_size']==3]['pulse_height'], bins=bins)
    axs[1, 0].set_title('Plateau size = 3')
    axs[1, 1].hist(df.loc[df['plateau_size']==4]['pulse_height'], bins=bins)
    axs[1, 1].set_title('Plateau size = 4')
    axs[2, 0].hist(df.loc[df['plateau_size']==4]['pulse_height'], bins=bins)
    axs[2, 0].set_title('Plateau size = 4')
    axs[2, 1].hist(df.loc[df['plateau_size']==5]['pulse_height'], bins=bins)
    axs[2, 1].set_title('Plateau size = 5')

    axs[0,0].yaxis.set_major_formatter(mtick.FormatStrFormatter('%.1e'))
    axs[0,1].yaxis.set_major_formatter(mtick.FormatStrFormatter('%.1e'))
    axs[1,0].yaxis.set_major_formatter(mtick.FormatStrFormatter('%.1e'))
    axs[1,1].yaxis.set_major_formatter(mtick.FormatStrFormatter('%.1e'))

    for ax in axs.flat:

        ax.set(xlabel='Pulse height', ylabel='count')

    plt.tight_layout()
    plt.savefig(img_save_path+'shaula_calib_ch0_EDA_plateau.png')
    
def skew_kurt(df, plat_size=1, img_save_path=None):
    df_1 = df.loc[df['plateau_size']==plat_size]
    sns.histplot(
        df_1, x="skewness", y="kurtosis",
        bins=30, discrete=(False, False), log_scale=(False, True),
        cbar=True, cbar_kws=dict(shrink=.97),
    )
    plt.savefig(f"{img_save_path}calib_data_skew_vs_kurtosis_plat_size_{plat_size}.png")




def shape_filt(df,uniq_shape,platau_removal=True):
    '''
    This function removes shapes with platau if platau_removal is True(Bydefault)
    if set to False then it simply removes duplicates from uniq_shape
    It takes shapes extracted from calibration measurements in the format of numpy
    array, and created dataframe from "pulse_data" function
    '''
    if platau_removal is True:
        two = df.loc[df['plateau_size']==2]['pulse_idx'].to_list()
        three = df.loc[df['plateau_size']==3]['pulse_idx'].to_list()
        four = df.loc[df['plateau_size']==4]['pulse_idx'].to_list()
        five = df.loc[df['plateau_size']==5]['pulse_idx'].to_list()
        removal_idx = two+three+four+five
        platauless_shape = np.delete(uniq_shape, [removal_idx], axis=0)
        
        shapes = [platauless_shape[i][platauless_shape[i]!=0] for i in range(platauless_shape.shape[0])]# creates list of shapes
        uniq_shapes = [list(j) for j in set(map(tuple,shapes))]# sublist match in the form of tuple
        return platauless_shape, uniq_shapes
    else:
        shapes = [uniq_shape[i][uniq_shape[i]!=0] for i in range(uniq_shape.shape[0])]# creates list of shapes
        uniq_shapes = [list(j) for j in set(map(tuple,shapes))]# sublist match in the form of tuple
        return uniq_shape
    

def filter_shapes_based_on_plat_size(df, uniq_shape, platau_removal=True, lst=[2,3,4,5], save_path=None):
    '''
    This function removes shapes with platau if platau_removal is True(Bydefault)
    if set to False then it simply removes duplicates from uniq_shape
    It takes shapes extracted from calibration measurements in the format of numpy
    array, and created dataframe from "pulse_data" function
    '''
    df1 = df.copy()
    if platau_removal is True:
        ind_lst = [df1.loc[df1['plateau_size']==i]['pulse_idx'].to_list() for i in lst]
        concat_ind_lst = [j for i in ind_lst for j in i]
        platauless_shape = np.delete(uniq_shape, concat_ind_lst, axis=0)
        save_name1 = ''.join([str(k) for k in lst])
        save_name0 = 'plateauless_uniq_pulses_'
        np.save(save_path + save_name0 + save_name1 + '.npy', platauless_shape)
        return platauless_shape, uniq_shape
    else:
        #shapes = [uniq_shape[i][uniq_shape[i]!=0] for i in range(uniq_shape.shape[0])]# creates list of shapes
        #uniq_shapes = [list(j) for j in set(map(tuple,shapes))]# sublist match in the form of tuple
        return uniq_shape
    
    
    
   
    
def get_consecutive_numbers(num, n, m):
    '''
    get n numbers before and after a provided number with a difference of m 
    '''
    return list(range(num-n*m, num+(n+1)*m, m))

def replace_with_unique_element(arr, n=1):
    """
    Randomly selects a unique element from a list or NumPy array `arr`,
    selects `n` indices in `arr` (excluding the index of the selected unique
    element), and replaces the values at these selected indices with the selected
    unique element.
    """
    # Select unique element
    unique_element = np.random.choice(np.unique(arr))

    # Select indices to replace
    indices_to_replace = np.random.choice(np.delete(np.arange(len(arr)), np.where(arr == unique_element)), size=n, replace=False)

    # Replace values at selected indices with unique element
    arr[indices_to_replace] = unique_element

    return arr
    
def shape_imposer(X,Y,S,l_a,l_b,POS):
    D_x = []
    D_y = []
    for x,y in zip(X,Y):
        #set equal no of samples as of pos randomly
        shape_selection = np.random.randint(0,len(S),len((POS)))
        shapes = [S[i] for i in shape_selection]
        L_a = [l_a[j] for j in shape_selection]
        L_b = [l_b[k] for k in shape_selection]
        # Now that the length of L_a,L_b, shapes and pos are same, we can use for loop for data generation
        for s, la, lb, po in zip(shapes,L_a,L_b,POS):
            pks, _ = find_peaks(s)
            if len(pks)!=0:
                b = len(s[:pks[0]])
                a = len(s[pks[0]:])
                #print(s)
                #print(lb)
                #print(po)
                x[po-b:po]+=lb
                x[po:po+a]+=la
                y[po]+=1
            elif len(pks)==0:
                b = 0
                a = len(s)
                x[po:po+len(s)]+=s
                y[po-b+1]+=1            
        D_x.append(x)
        D_y.append(y)
    return D_x,D_y

def controlled_mc(shape_file, required_examples, sample_size, indice_to_add):
    with open(shape_file, 'rb') as f:
        S = pickle.load(f)
    l_a = []
    l_b = []
    for i in S:
        sh = i
        pk,_ = find_peaks(i)
        if len(pk)==0:
            peak = 0
            la = len(i)
            lb = 0
        else:
            peak = pk[0]
            lb = i[:peak]
            la = i[peak:]
        l_a.append(la)
        l_b.append(lb)
    X = np.zeros((required_examples, sample_size))
    Y = np.zeros((required_examples, sample_size))
    #pos = [85,90,95,100,105,110,115,120]
    x, y = shape_imposer(X,Y,S,l_a,l_b,indice_to_add)
    return X, Y


def prediction_error(ch0_path, ch0_label_path, model_dir_path_form_glob, examples):
    for i in glob.glob(model_dir_path_form_glob)[:]:
        print('*'*50)
        print(i)
        print('*'*50)

        test_x = np.load(ch0_path)
        test_y = np.load(ch0_label_path)

        model = keras.models.load_model(i,compile=False)
        model.compile()
        predictions = model.predict(test_x)

        #import matplotlib.pyplot as plt
        #import numpy as np
        from matplotlib import gridspec
        
        samp_ind = np.random.randint(examples, test_x.shape[0],1)[0]#--->268mhz

        fig = plt.figure(figsize=(17,5))
        # set height ratios for subplots
        gs = gridspec.GridSpec(2, 1, height_ratios=[2, 1]) 
        # the first subplot
        ax0 = plt.subplot(gs[0])
        # log scale for axis Y of the first subplot
        #ax0.set_yscale("log")
        line0, = ax0.plot(test_x[samp_ind], color='black')
        ax0.set_ylabel('Counts')
        # the second subplot
        # shared axis X
        ax1 = plt.subplot(gs[1], sharex = ax0)
        line1, = ax1.plot(test_y[samp_ind], color='b', marker='o', linestyle='-')
        line2, = ax1.plot(predictions[samp_ind], color='red', marker='.', linestyle='-', alpha=0.9)
        plt.setp(ax0.get_xticklabels(), visible=False)
        ax1.set_ylabel('Photons')
        # remove last tick label for the second subplot
        yticks = ax1.yaxis.get_major_ticks()
        yticks[-1].label1.set_visible(False)
        # put legend on first subplot
        ax0.legend((line0, line1, line2), ('Data sample', 'Label vector', 'Predictions'), loc='upper left')
        # remove vertical gap between subplots
        plt.subplots_adjust(hspace=.0)
        ax0.set_title('Ground truth({}MHz) and prediction({}MHz) comparison'.format(np.round((1000/(1.6*256))*np.sum(test_y[samp_ind]),2), np.round((1000/(1.6*256))*np.sum(predictions[samp_ind]),2)))
        ax1.set_xlabel('Time-step (1.6ns)')
        #plt.savefig('Label({}MHz)-pred_comp({}MHz)'.format(np.int((1000/(1.6*256))*np.sum(test_y[samp_ind])), np.int((1000/(1.6*256))*np.sum(predictions[samp_ind]))))
        plt.show()

        avg_true_rate = []
        avg_pred_rate = []
        std_pred_rate = []

        flat_predictions = []
        flat_labels = []

        cluster_pred = []
        cluster_labels=[]

        predictions1 = predictions.reshape(predictions.shape[0],predictions.shape[1])
        test_y1 = test_y.reshape(test_y.shape[0],test_y.shape[1])
        ref_arr = np.arange(predictions.shape[0])[::int(predictions.shape[0]/predictions.shape[1])]
        for i in range(1,len(ref_arr))[:]:
            low_lim = ref_arr[i-1]
            up_lim = ref_arr[i]

            pred_set = predictions1[low_lim:up_lim]
            label_set = test_y1[low_lim:up_lim]
            flat_pred = pred_set.flatten()
            flat_pred = np.rint(flat_pred)
            flat_pred_original = np.array([iii for iii in flat_pred])
            flat_label = label_set.flatten()

            #print(flat_pred)
            clus_pred = np.diff(flat_pred)
            clus_label = np.diff(flat_label)

            flat_predictions.append(flat_pred)
            flat_labels.append(flat_label)

            cluster_pred.append(clus_pred)
            cluster_labels.append(clus_label)

            #
            true_rate = (1000/(1.6*256))*np.mean(np.sum(label_set,axis=1))
            avg_true_rate.append(true_rate)

            pred_rate = (1000/(1.6*256))*np.mean(np.sum(pred_set,axis=1))
            avg_pred_rate.append(pred_rate)

            pred_rate_std = (1000/(1.6*256))*np.std(np.sum(label_set - pred_set,axis=1))
            std_pred_rate.append(pred_rate_std)

        flat_predictions = np.array(flat_predictions)
        flat_labels = np.array(flat_labels)

        cluster_pred = np.array(cluster_pred)
        cluster_labels = np.array(cluster_labels)

        avg_true_rate = np.array(avg_true_rate)
        avg_pred_rate = np.array(avg_pred_rate)
        std_pred_rate = np.array(std_pred_rate)

        error = avg_true_rate - avg_pred_rate

        import matplotlib as mpl

        # Select a color map
        cmap = mpl.cm.viridis#bwr

        # Some Test data

        x = avg_true_rate # np.linspace(-4, 4, npts) # avg_true_rate
        y = error # norm.pdf(x) # error
        z = std_pred_rate # np.sin(2 * x) # std_pred_rate
        normalize = mpl.colors.Normalize(vmin=z.min(), vmax=z.max())

        # The plot
        fig = plt.figure(figsize=(18,7))
        ax = fig.add_axes([0.12, 0.12, 0.68, 0.78])
        plt.plot(x, y, color="black")
        plt.xlabel('Rate (MHz)')
        plt.ylabel(' Average error (MHz)')
        for i in range(len(x)-1):
            plt.fill_between([x[i], x[i+1]], [y[i], y[i+1]], color=cmap(normalize(z[i])))

        cbax = fig.add_axes([0.82, 0.12, 0.03, 0.78])
        cb = mpl.colorbar.ColorbarBase(cbax, cmap=cmap, norm=normalize, orientation='vertical')
        cb.set_label("Error uncertainty in a waveform (MHz)", rotation=270, labelpad=15)
        #plt.savefig('shaula_00000-003_ch0_error_ana.png')

def confusion_mat(ch0_path, ch0_label_path, model_dir_path_form_glob):
    for i in glob.glob(model_dir_path_form_glob):
        print(i)
        test_x = np.load(ch0_path)
        test_y = np.load(ch0_label_path)
        model = keras.models.load_model(i,compile=False)
        model.compile()

        predictions = model.predict(test_x)

        lower_lim = 6
        upper_lim = test_y.shape[1]-10

        test_y1 = test_y[:, lower_lim:upper_lim]
        predictions1 = predictions.reshape(predictions.shape[0],predictions.shape[1])[:, lower_lim:upper_lim]
        predictions1 = np.rint(predictions1)# round the predictions
        test_y1 = np.rint(test_y1)# riund the labels


        cm = confusion_matrix(test_y1.ravel(), predictions1.ravel())
        cm = cm/cm.sum(axis=1)[:,None]

        where_are_NaNs = np.isnan(cm)
        cm[where_are_NaNs] = 0
        cm = np.around(cm, decimals=2)
        cm_display = metrics.ConfusionMatrixDisplay(confusion_matrix = cm)
        #cm_display.plot()
        fig, ax = plt.subplots(figsize=(13,10))
        cm_display.plot(ax=ax)
        #plt.savefig('recall_confusion_matrix.png')
        plt.show()
import numpy as np

def remove_nonzero_neighbors(arr):
    # Find the indices of the smallest elements in the array
    smallest_indices = np.where(arr == np.min(arr))[0]

    # Remove the indices of smallest elements whose neighbors are non-zero
    modified_indices = []
    for i in smallest_indices:
        if i > 0 and i < len(arr) - 1:
            if arr[i-1] == 0 and arr[i+1] == 0:
                modified_indices.append(i)

    return modified_indices



import numpy as np
import pandas as pd
import pickle
import sys
sys.path.append("..")
import iimcsim.data_gen as data_gen
import iimcsim.tools as fg

# Required files: MC.py, monte_carlo.py and pulse shape file in 
# .npy format(here plateauless_uniq_pulses.npy)

def generate_data(train_exe = 50,
                name = './src/data/plateauless_uniq_pulses_345.npy',
                samp_size = 256,
                max_num_photon = 60,
                folder_name = 'sh06x2'):
    data_gen.generator(samp_size, max_num_photon, name, train_exe, folder_name)
#generate_data()


import numpy as np
import pandas as pd
import pickle
from iimcsim.MonteCarlo import MonteCarlo
from iimcsim.MC import MC_set
from iimcsim.shape_ext import shape_generator
import os

# Required files: MC.py, monte_carlo.py and pulse shape file in 
# .npy format(here plateauless_uniq_pulses.npy)

def generator(samp_size, max_rate, name, train_exe, folder_name = 'sh06x2'):
    
    '''
    This function generates MC data for training, testing and validation,
    and save them to "./DATASETS/sub_folder_name/" path.
    
    All dataset are generated using given pulse shapes, unlike separate splitting
    of set of pulse shapes for train, test and val data generation.
    
    Test and val data has similar size(however can be changed in this function)
    The ratio between training dataset size and validation dataset size is 0.33 here.
    
    All datasets has specific format: range of rate(in photon number) with desired
    number of sample examples for each rate.
    '''
    
    name1 = name.split('.')[0]
    
    val_exe = test_exe = int(train_exe*0.33)
    
    
    x = MC_set(name, max_rate, train_exe, samp_size)
    remove_duplicates = x.remove_duplicate_pulse()
    train,evaluation = x.train_test_val_split(60,40)
    train_x, train_y, train_x1, train_y1, CORR_pattern, CH0_ind, CH1_ind = x.get_mc_set(remove_duplicates)
    
    path = f'./DATASETS/{folder_name}/'
    # Check whether the specified path exists or not
    isExist = os.path.exists(path)
    if not isExist:
        # Create a new directory because it does not exist
        os.makedirs(path)
        #print("The new directory is created!")

    np.save(f'./DATASETS/{folder_name}/lab_ch0_shapes_train_x_m_{train_exe}.npy',train_x)
    np.save(f'./DATASETS/{folder_name}/lab_ch0_shapes_train_y_m_{train_exe}.npy',train_y)
    
    np.save(f'./DATASETS/{folder_name}/lab_ch1_shapes_train_x_m_{train_exe}.npy',train_x1)
    np.save(f'./DATASETS/{folder_name}/lab_ch1_shapes_train_y_m_{train_exe}.npy',train_y1)
    
    with open(f'./DATASETS/{folder_name}/lab_ch0_1_shapes_train_CORR_pattern_{train_exe}', 'wb') as fp:
        pickle.dump(CORR_pattern, fp)
    with open(f'./DATASETS/{folder_name}/lab_shapes_train_CH0_ind_{train_exe}', 'wb') as ch0p:
        pickle.dump(CH0_ind, ch0p)
    with open(f'./DATASETS/{folder_name}/lab_shapes_train_CH1_ind_{train_exe}', 'wb') as ch1p:
        pickle.dump(CH1_ind, ch1p) 
        
    
    x = MC_set(name, max_rate, test_exe, samp_size)
    remove_duplicates = x.remove_duplicate_pulse()
    train,evaluation = x.train_test_val_split(60,40)
    test_x, test_y, test_x1, test_y1, CORR_pattern, CH0_ind, CH1_ind = x.get_mc_set(remove_duplicates)
    np.save(f'./DATASETS/{folder_name}/lab_ch0_shapes_test_x_m_{train_exe}.npy',test_x)
    np.save(f'./DATASETS/{folder_name}/lab_ch0_shapes_test_y_m_{train_exe}.npy',test_y)
    
    np.save(f'./DATASETS/{folder_name}/lab_ch1_shapes_test_x_m_{train_exe}.npy',test_x1)
    np.save(f'./DATASETS/{folder_name}/lab_ch1_shapes_test_y_m_{train_exe}.npy',test_y1)
    
    with open(f'./DATASETS/{folder_name}/lab_ch0_1_shapes_test_CORR_pattern_{train_exe}', 'wb') as fp:
        pickle.dump(CORR_pattern, fp)
    with open(f'./DATASETS/{folder_name}/lab_shapes_test_CH0_ind_{train_exe}', 'wb') as ch0p:
        pickle.dump(CH0_ind, ch0p)
    with open(f'./DATASETS/{folder_name}/lab_shapes_test_CH1_ind_{train_exe}', 'wb') as ch1p:
        pickle.dump(CH1_ind, ch1p)
    
    
    x = MC_set(name, max_rate, val_exe, samp_size)
    remove_duplicates = x.remove_duplicate_pulse()
    train,evaluation = x.train_test_val_split(60,40)
    val_x, val_y, val_x1, val_y1, CORR_pattern, CH0_ind, CH1_ind = x.get_mc_set(remove_duplicates)
    np.save(f'./DATASETS/{folder_name}/lab_ch0_shapes_val_x_m_{train_exe}.npy',val_x)
    np.save(f'./DATASETS/{folder_name}/lab_ch0_shapes_val_y_m_{train_exe}.npy',val_y)
    
    np.save(f'./DATASETS/{folder_name}/lab_ch1_shapes_val_x_m_{train_exe}.npy',val_x1)
    np.save(f'./DATASETS/{folder_name}/lab_ch1_shapes_val_y_m_{train_exe}.npy',val_y1)
    
    with open(f'./DATASETS/{folder_name}/lab_ch0_1_shapes_val_CORR_pattern_{train_exe}', 'wb') as fp:
        pickle.dump(CORR_pattern, fp)
    with open(f'./DATASETS/{folder_name}/lab_shapes_val_CH0_ind_{train_exe}', 'wb') as ch0p:
        pickle.dump(CH0_ind, ch0p)
    with open(f'./DATASETS/{folder_name}/lab_shapes_val_CH1_ind_{train_exe}', 'wb') as ch1p:
        pickle.dump(CH1_ind, ch1p)
        
        
    
    print(train_x.shape,val_x.shape,test_x.shape)
    print(train_y.shape,val_y.shape,test_y.shape)
